try:
    import simplejson as json
except ImportError:
    try:
        import rapidjson as json
    except ImportError:
        try:
            import ujson as json
        except ImportError:
            import json



from attr import attrs, attrib

from aioalice.types import AliceObject, MediaButton
from aioalice.utils import ensure_cls


@attrs
class Image(AliceObject):
    """Image object"""
    image_id = attrib(type=str)
    title = attrib(type=str)
    description = attrib(type=str)
    button = attrib(default=None, converter=ensure_cls(MediaButton))



import asyncio

from aioalice.dispatcher import Dispatcher
from aioalice.dispatcher.webhook import get_new_configured_app, configure_app

try:
    import uvloop
except ImportError:
    pass
else:
    asyncio.set_event_loop_policy(uvloop.EventLoopPolicy())


__version__ = '1.5.0'



"""
https://gist.github.com/mahenzon/a6c2804a2d18a2ab75630bb5d93693c8
"""

import functools
from inspect import isclass, getfullargspec


def safe_kwargs(func_or_class):
    from aioalice.types.base import AliceObject

    spec = getfullargspec(func_or_class)
    all_args = spec.args

    save_raw_kwargs = isclass(func_or_class) and issubclass(func_or_class, AliceObject)

    @functools.wraps(func_or_class)
    def wrap(*args, **kwargs):
        accepted_kwargs = {k: v for k, v in kwargs.items() if k in all_args}
        res = func_or_class(*args, **accepted_kwargs)

        if save_raw_kwargs:
            # saving all kwargs for access to unexpected attrs
            res._raw_kwargs.update(kwargs)

        return res

    return wrap



from aiohttp.web import Request as WebRequest
from attr import attrs, attrib

from aioalice.types import (
    AliceObject,
    Meta,
    Session,
    Card,
    Request,
    Response,
    AliceResponse,
)
from aioalice.utils import ensure_cls, safe_kwargs


@safe_kwargs
@attrs
class AliceRequest(AliceObject):
    """AliceRequest is a request from Alice API"""
    original_request = attrib(type=WebRequest)
    meta = attrib(converter=ensure_cls(Meta))
    request = attrib(converter=ensure_cls(Request))
    session = attrib(converter=ensure_cls(Session))
    version = attrib(type=str)

    def _response(self, response, session_state=None, user_state_update=None, application_state=None):
        return AliceResponse(
            response=response,
            session=self.session.base,
            session_state=session_state or {},
            user_state_update=user_state_update or {},
            application_state=application_state or {},
            version=self.version,
        )

    def response(self, response_or_text, session_state=None, user_state_update=None, application_state=None, **kwargs):
        """
        Generate response

        :param response_or_text: Response or Response's text:
            if response_or_text is not an instance of Response,
            it is passed to the Response initialisator with kwargs.
            Otherwise it is used as a Response

        :param kwargs: tts, card, buttons, end_session for Response
            NOTE: if you want to pass card, consider using one of
              these methods: response_big_image, response_items_list

        :param session_state: Session's state
        :param user_state_update: User's state
        :param application_state: Application's state
            Allows to store data on Yandex's side
            Read more: https://yandex.ru/dev/dialogs/alice/doc/session-persistence.html

        :return: AliceResponse
        """
        if not isinstance(response_or_text, Response):
            response_or_text = Response(response_or_text, **kwargs)
        return self._response(response_or_text, session_state, user_state_update, application_state)

    def response_big_image(self, text, image_id, title, description, button=None,
                           session_state=None, user_state_update=None, application_state=None, **kwargs):
        """
        Generate response with Big Image card

        :param text: Response's text
        :param image_id: Image's id for BigImage Card
        :param title: Image's title for BigImage Card
        :param description: Image's description for BigImage Card
        :param button: Image's button for BigImage Card
        :param session_state: Session's state
        :param user_state_update: User's state
        :param application_state: Application's state
            Allows to store data on Yandex's side
            Read more: https://yandex.ru/dev/dialogs/alice/doc/session-persistence.html
        :param kwargs: tts, buttons, end_session for Response
        :return: AliceResponse
        """
        return self._response(
            Response(
                text,
                card=Card.big_image(image_id, title, description, button),
                **kwargs,
            ),
            session_state, user_state_update, application_state
        )

    def response_items_list(self, text, header, items, footer=None,
                            session_state=None, user_state_update=None, application_state=None, **kwargs):
        """
        Generate response with Items List card

        :param text: Response's text
        :param header: Card's header
        :param items: Card's items - list of `Image` objects
        :param footer: Card's footer
        :param session_state: Session's state
        :param user_state_update: User's state
        :param application_state: Application's state
            Allows to store data on Yandex's side
            Read more: https://yandex.ru/dev/dialogs/alice/doc/session-persistence.html
        :param kwargs: tts, buttons, end_session for Response
        :return: AliceResponse
        """
        return self._response(
            Response(
                text,
                card=Card.items_list(header, items, footer),
                **kwargs,
            ),
            session_state, user_state_update, application_state
        )



from attr import attrs, attrib, asdict


@attrs
class AliceObject:
    """AliceObject is base class for all Alice requests related objects"""

    _raw_kwargs = attrib(factory=dict, init=False)
    """
    here the raw JSON (dict) will be stored
    for using with compatible API
    """

    def to_json(self):
        data = asdict(self, filter=filter_to_json)
        return data


def filter_to_json(attr, value) -> bool:
    return attr.name != '_raw_kwargs'



from attr import attrs, attrib

from aioalice.types import AliceObject


@attrs
class Interfaces(AliceObject):
    """Interfaces object"""
    account_linking = attrib(factory=dict)
    payments = attrib(factory=dict)
    screen = attrib(factory=dict)



from aioalice.utils import exceptions
from aioalice.utils.json import json
from aioalice.utils.payload import generate_json_payload
from aioalice.utils.safe_kwargs import safe_kwargs


def ensure_cls(klass):
    from aioalice.types.base import AliceObject
    safe_cls = safe_kwargs(klass) if issubclass(klass, AliceObject) else klass

    def converter(val):
        if val is None:
            return
        if isinstance(val, dict):
            return safe_cls(**val)
        if isinstance(val, list):
            return [converter(v) for v in val]
        if not isinstance(val, klass):
            return klass(val)
        return val
    return converter


__all__ = [
    "exceptions",
    "json",
    "generate_json_payload",
    "safe_kwargs",
    "ensure_cls",
]



from aioalice.types.base import AliceObject
from aioalice.types.interfaces import Interfaces
from aioalice.types.meta import Meta
from aioalice.types.markup import Markup
from aioalice.types.entity_tokens import EntityTokens
from aioalice.types.entity_value import EntityValue
from aioalice.types.entity import Entity, EntityType
from aioalice.types.natural_language_understanding import NaturalLanguageUnderstanding
from aioalice.types.request import Request, RequestType
from aioalice.types.session import BaseSession, Session

from aioalice.types.quota import Quota
from aioalice.types.uploaded_image import UploadedImage
from aioalice.types.media_button import MediaButton
from aioalice.types.image import Image
from aioalice.types.card_header import CardHeader
from aioalice.types.card_footer import CardFooter
from aioalice.types.card import Card, CardType

from aioalice.types.button import Button
from aioalice.types.response import Response

from aioalice.types.alice_response import AliceResponse
from aioalice.types.alice_request import AliceRequest



from attr import attrs, attrib

from aioalice.types import AliceObject, BaseSession, Response
from aioalice.utils import ensure_cls


@attrs
class AliceResponse(AliceObject):
    """AliceResponse is a response to Alice API"""

    response = attrib(converter=ensure_cls(Response))
    session = attrib(converter=ensure_cls(BaseSession))
    session_state = attrib(type=dict)
    user_state_update = attrib(type=dict)
    application_state = attrib(type=dict)
    version = attrib(type=str)



from attr import attrs, attrib

from aioalice.types import AliceObject
from aioalice.utils import safe_kwargs


@safe_kwargs
@attrs
class Markup(AliceObject):
    """Markup object"""
    dangerous_context = attrib(type=bool)
